@@ -22,7 +22,7 @@
  *
  *   You should have received a copy of the GNU Lesser General Public
  *   License along with this library; if not, write to the Free Software
- *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
   
@@ -172,6 +172,8 @@ int snd_ext_parm_mask_refine(snd_mask_t *mask, struct snd_ext_parm *parm, int ty
 	unsigned int i;
 
 	parm += type;
+	if (!parm->active)
+		return 0;
 	memset(&bits, 0, sizeof(bits));
 	for (i = 0; i < parm->num_list; i++)
 		bits.bits[parm->list[i] / 32] |= 1U << (parm->list[i] % 32);
@@ -247,7 +249,7 @@ static unsigned int get_links(struct snd_ext_parm *params)
 			      SND_PCM_HW_PARBIT_TICK_TIME);
 
 	for (i = 0; i < SND_PCM_EXTPLUG_HW_PARAMS; i++) {
-		if (params[i].active)
+		if (params[i].active && !params[i].keep_link)
 			links &= ~excl_parbits[i];
 	}
 	return links;
@@ -425,6 +427,36 @@ static int snd_pcm_extplug_close(snd_pcm_t *pcm)
 	return 0;
 }
 
+static snd_pcm_chmap_query_t **snd_pcm_extplug_query_chmaps(snd_pcm_t *pcm)
+{
+	extplug_priv_t *ext = pcm->private_data;
+
+	if (ext->data->version >= 0x010002 &&
+	    ext->data->callback->query_chmaps)
+		return ext->data->callback->query_chmaps(ext->data);
+	return snd_pcm_generic_query_chmaps(pcm);
+}
+
+static snd_pcm_chmap_t *snd_pcm_extplug_get_chmap(snd_pcm_t *pcm)
+{
+	extplug_priv_t *ext = pcm->private_data;
+
+	if (ext->data->version >= 0x010002 &&
+	    ext->data->callback->get_chmap)
+		return ext->data->callback->get_chmap(ext->data);
+	return snd_pcm_generic_get_chmap(pcm);
+}
+
+static int snd_pcm_extplug_set_chmap(snd_pcm_t *pcm, const snd_pcm_chmap_t *map)
+{
+	extplug_priv_t *ext = pcm->private_data;
+
+	if (ext->data->version >= 0x010002 &&
+	    ext->data->callback->set_chmap)
+		return ext->data->callback->set_chmap(ext->data, map);
+	return snd_pcm_generic_set_chmap(pcm, map);
+}
+
 static const snd_pcm_ops_t snd_pcm_extplug_ops = {
 	.close = snd_pcm_extplug_close,
 	.info = snd_pcm_generic_info,
@@ -438,6 +470,9 @@ static const snd_pcm_ops_t snd_pcm_extplug_ops = {
 	.async = snd_pcm_generic_async,
 	.mmap = snd_pcm_generic_mmap,
 	.munmap = snd_pcm_generic_munmap,
+	.query_chmaps = snd_pcm_extplug_query_chmaps,
+	.get_chmap = snd_pcm_extplug_get_chmap,
+	.set_chmap = snd_pcm_extplug_set_chmap,
 };
 
 #endif /* !DOC_HIDDEN */
@@ -607,6 +642,17 @@ as former functions.
 To clear the parameter constraints, call #snd_pcm_extplug_params_reset()
 function. 
 
+When using snd_pcm_extplug_set_param_*() or snd_pcm_extplug_set_slave_param_*()
+for any parameter. This parameter is no longer linked between the client and
+slave PCM. Therefore it could differ and the extplug has to support conversion
+between all valid parameter configurations. To keep the client and slave
+parameter linked #snd_pcm_extplug_set_param_link() can be used for the
+corresponding parameter. For example if the extplug does not support channel nor
+format conversion the supported client parameters can be limited with
+snd_pcm_extplug_set_param_*() and afterwards
+#snd_pcm_extplug_set_param_link(ext, SND_PCM_EXTPLUG_HW_FORMAT, 1) and
+#snd_pcm_extplug_set_param_link(ext, SND_PCM_EXTPLUG_HW_CHANNELS, 1) should be
+called to keep the client and slave parameters the same.
 */
 
 /**
@@ -686,6 +732,7 @@ int snd_pcm_extplug_create(snd_pcm_extplug_t *extplug, const char *name,
 	pcm->private_data = ext;
 	pcm->poll_fd = spcm->poll_fd;
 	pcm->poll_events = spcm->poll_events;
+	pcm->tstamp_type = spcm->tstamp_type;
 	snd_pcm_set_hw_ptr(pcm, &ext->plug.hw_ptr, -1, 0);
 	snd_pcm_set_appl_ptr(pcm, &ext->plug.appl_ptr, -1, 0);
 
@@ -733,7 +780,7 @@ void snd_pcm_extplug_params_reset(snd_pcm_extplug_t *extplug)
 int snd_pcm_extplug_set_slave_param_list(snd_pcm_extplug_t *extplug, int type, unsigned int num_list, const unsigned int *list)
 {
 	extplug_priv_t *ext = extplug->pcm->private_data;
-	if (type < 0 && type >= SND_PCM_EXTPLUG_HW_PARAMS) {
+	if (type < 0 || type >= SND_PCM_EXTPLUG_HW_PARAMS) {
 		SNDERR("EXTPLUG: invalid parameter type %d", type);
 		return -EINVAL;
 	}
@@ -755,7 +802,7 @@ int snd_pcm_extplug_set_slave_param_list(snd_pcm_extplug_t *extplug, int type, u
 int snd_pcm_extplug_set_slave_param_minmax(snd_pcm_extplug_t *extplug, int type, unsigned int min, unsigned int max)
 {
 	extplug_priv_t *ext = extplug->pcm->private_data;
-	if (type < 0 && type >= SND_PCM_EXTPLUG_HW_PARAMS) {
+	if (type < 0 || type >= SND_PCM_EXTPLUG_HW_PARAMS) {
 		SNDERR("EXTPLUG: invalid parameter type %d", type);
 		return -EINVAL;
 	}
@@ -781,7 +828,7 @@ int snd_pcm_extplug_set_slave_param_minmax(snd_pcm_extplug_t *extplug, int type,
 int snd_pcm_extplug_set_param_list(snd_pcm_extplug_t *extplug, int type, unsigned int num_list, const unsigned int *list)
 {
 	extplug_priv_t *ext = extplug->pcm->private_data;
-	if (type < 0 && type >= SND_PCM_EXTPLUG_HW_PARAMS) {
+	if (type < 0 || type >= SND_PCM_EXTPLUG_HW_PARAMS) {
 		SNDERR("EXTPLUG: invalid parameter type %d", type);
 		return -EINVAL;
 	}
@@ -803,7 +850,7 @@ int snd_pcm_extplug_set_param_list(snd_pcm_extplug_t *extplug, int type, unsigne
 int snd_pcm_extplug_set_param_minmax(snd_pcm_extplug_t *extplug, int type, unsigned int min, unsigned int max)
 {
 	extplug_priv_t *ext = extplug->pcm->private_data;
-	if (type < 0 && type >= SND_PCM_EXTPLUG_HW_PARAMS) {
+	if (type < 0 || type >= SND_PCM_EXTPLUG_HW_PARAMS) {
 		SNDERR("EXTPLUG: invalid parameter type %d", type);
 		return -EINVAL;
 	}
@@ -814,3 +861,26 @@ int snd_pcm_extplug_set_param_minmax(snd_pcm_extplug_t *extplug, int type, unsig
 	return snd_ext_parm_set_minmax(&ext->params[type], min, max);
 }
 
+/**
+ * @brief Keep the client and slave format/channels the same if requested. This
+ * is for example useful if this extplug does not support any channel
+ * conversion.
+ * @param extplug the extplug handle
+ * @param type parameter type
+ * @param keep_link if 1 the parameter identified by type will be kept the same
+ * for the client and slave PCM of this extplug
+ * @return 0 if successful, or a negative error code
+ */
+int snd_pcm_extplug_set_param_link(snd_pcm_extplug_t *extplug, int type,
+				   int keep_link)
+{
+	extplug_priv_t *ext = extplug->pcm->private_data;
+
+	if (type < 0 || type >= SND_PCM_EXTPLUG_HW_PARAMS) {
+		SNDERR("EXTPLUG: invalid parameter type %d", type);
+		return -EINVAL;
+	}
+	ext->params[type].keep_link = keep_link ? 1 : 0;
+	ext->sparams[type].keep_link = keep_link ? 1 : 0;
+	return 0;
+}
