@@ -11,7 +11,7 @@
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software  
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
  *  Support for the verb/device/modifier core logic and API,
  *  command line tool and file parser was kindly sponsored by
@@ -34,28 +34,46 @@
 #include <ctype.h>
 #include <stdarg.h>
 #include <pthread.h>
+#include <sys/stat.h>
+#include <limits.h>
 
 /*
  * misc
  */
 
-static int get_value1(const char **value, struct list_head *value_list,
-                      const char *identifier);
-static int get_value3(const char **value,
+static int get_value(snd_use_case_mgr_t *uc_mgr,
+			const char *identifier,
+			char **value,
+			const char *mod_dev_name,
+			const char *verb_name,
+			int exact);
+static int get_value1(snd_use_case_mgr_t *uc_mgr, char **value,
+                      struct list_head *value_list, const char *identifier);
+static int get_value3(snd_use_case_mgr_t *uc_mgr,
+                      char **value,
 		      const char *identifier,
 		      struct list_head *value_list1,
 		      struct list_head *value_list2,
 		      struct list_head *value_list3);
 
+static int execute_component_seq(snd_use_case_mgr_t *uc_mgr,
+				 struct component_sequence *cmpt_seq,
+				 struct list_head *value_list1,
+				 struct list_head *value_list2,
+				 struct list_head *value_list3,
+				 char *cdev);
+
 static int check_identifier(const char *identifier, const char *prefix)
 {
 	int len;
 
-	if (strcmp(identifier, prefix) == 0)
-		return 1;
 	len = strlen(prefix);
-	if (memcmp(identifier, prefix, len) == 0 && identifier[len] == '/')
+	if (strncmp(identifier, prefix, len) != 0)
+		return 0;
+
+	if (identifier[len] == 0 || identifier[len] == '/')
 		return 1;
+
 	return 0;
 }
 
@@ -129,47 +147,123 @@ int snd_use_case_free_list(const char *list[], int items)
 	return 0;
 }
 
-static int open_ctl(snd_use_case_mgr_t *uc_mgr,
-		    snd_ctl_t **ctl,
-		    const char *ctl_dev)
+static int read_tlv_file(unsigned int **res,
+			 const char *filepath)
 {
-	int err;
+	int err = 0;
+	int fd;
+	struct stat st;
+	size_t sz;
+	ssize_t sz_read;
+	struct snd_ctl_tlv *tlv;
 
-	/* FIXME: add a list of ctl devices to uc_mgr structure and
-           cache accesses for multiple opened ctl devices */
-	if (uc_mgr->ctl_dev != NULL && strcmp(ctl_dev, uc_mgr->ctl_dev) == 0) {
-		*ctl = uc_mgr->ctl;
-		return 0;
-	}
-	if (uc_mgr->ctl_dev) {
-		free(uc_mgr->ctl_dev);
-		uc_mgr->ctl_dev = NULL;
-		snd_ctl_close(uc_mgr->ctl);
-	
-	}
-	err = snd_ctl_open(ctl, ctl_dev, 0);
-	if (err < 0)
+	fd = open(filepath, O_RDONLY);
+	if (fd < 0) {
+		err = -errno;
 		return err;
-	uc_mgr->ctl_dev = strdup(ctl_dev);
-	if (uc_mgr->ctl_dev == NULL) {
-		snd_ctl_close(*ctl);
-		return -ENOMEM;
 	}
-	uc_mgr->ctl = *ctl;
-	return 0;
+	if (fstat(fd, &st) == -1) {
+		err = -errno;
+		goto __fail;
+	}
+	sz = st.st_size;
+	if (sz > 16 * 1024 * 1024 || sz < 8 || sz % 4) {
+		uc_error("File size should be less than 16 MB "
+			 "and multiple of 4");
+		err = -EINVAL;
+		goto __fail;
+	}
+	*res = malloc(sz);
+	if (res == NULL) {
+		err = -ENOMEM;
+		goto __fail;
+	}
+	sz_read = read(fd, *res, sz);
+	if (sz_read < 0 || (size_t)sz_read != sz) {
+		err = -EIO;
+		free(*res);
+		*res = NULL;
+	}
+	/* Check if the tlv file specifies valid size. */
+	tlv = (struct snd_ctl_tlv *)(*res);
+	if (tlv->length + 2 * sizeof(unsigned int) != sz) {
+		uc_error("Invalid tlv size: %d", tlv->length);
+		err = -EINVAL;
+		free(*res);
+		*res = NULL;
+	}
+
+__fail:
+	close(fd);
+	return err;
+}
+
+static int binary_file_parse(snd_ctl_elem_value_t *dst,
+			      snd_ctl_elem_info_t *info,
+			      const char *filepath)
+{
+	int err = 0;
+	int fd;
+	struct stat st;
+	size_t sz;
+	ssize_t sz_read;
+	char *res;
+	snd_ctl_elem_type_t type;
+	unsigned int idx, count;
+
+	type = snd_ctl_elem_info_get_type(info);
+	if (type != SND_CTL_ELEM_TYPE_BYTES) {
+		uc_error("only support byte type!");
+		err = -EINVAL;
+		return err;
+	}
+	fd = open(filepath, O_RDONLY);
+	if (fd < 0) {
+		err = -errno;
+		return err;
+	}
+	if (stat(filepath, &st) == -1) {
+		err = -errno;
+		goto __fail;
+	}
+	sz = st.st_size;
+	count = snd_ctl_elem_info_get_count(info);
+	if (sz != count || sz > sizeof(dst->value.bytes)) {
+		uc_error("invalid parameter size %d!", sz);
+		err = -EINVAL;
+		goto __fail;
+	}
+	res = malloc(sz);
+	if (res == NULL) {
+		err = -ENOMEM;
+		goto __fail;
+	}
+	sz_read = read(fd, res, sz);
+	if (sz_read < 0 || (size_t)sz_read != sz) {
+		err = -errno;
+		goto __fail_read;
+	}
+	for (idx = 0; idx < sz; idx++)
+		snd_ctl_elem_value_set_byte(dst, idx, *(res + idx));
+      __fail_read:
+	free(res);
+      __fail:
+	close(fd);
+	return err;
 }
 
 extern int __snd_ctl_ascii_elem_id_parse(snd_ctl_elem_id_t *dst,
 					 const char *str,
 					 const char **ret_ptr);
 
-static int execute_cset(snd_ctl_t *ctl, const char *cset)
+static int execute_cset(snd_ctl_t *ctl, const char *cset, unsigned int type)
 {
 	const char *pos;
 	int err;
 	snd_ctl_elem_id_t *id;
 	snd_ctl_elem_value_t *value;
 	snd_ctl_elem_info_t *info;
+	unsigned int *res = NULL;
 
 	snd_ctl_elem_id_malloc(&id);
 	snd_ctl_elem_value_malloc(&value);
@@ -185,20 +279,36 @@ static int execute_cset(snd_ctl_t *ctl, const char *cset)
 		err = -EINVAL;
 		goto __fail;
 	}
-	snd_ctl_elem_value_set_id(value, id);
 	snd_ctl_elem_info_set_id(info, id);
-	err = snd_ctl_elem_read(ctl, value);
-	if (err < 0)
-		goto __fail;
 	err = snd_ctl_elem_info(ctl, info);
 	if (err < 0)
 		goto __fail;
-	err = snd_ctl_ascii_value_parse(ctl, value, info, pos);
-	if (err < 0)
-		goto __fail;
-	err = snd_ctl_elem_write(ctl, value);
-	if (err < 0)
-		goto __fail;
+	if (type == SEQUENCE_ELEMENT_TYPE_CSET_TLV) {
+		if (!snd_ctl_elem_info_is_tlv_writable(info)) {
+			err = -EINVAL;
+			goto __fail;
+		}
+		err = read_tlv_file(&res, pos);
+		if (err < 0)
+			goto __fail;
+		err = snd_ctl_elem_tlv_write(ctl, id, res);
+		if (err < 0)
+			goto __fail;
+	} else {
+		snd_ctl_elem_value_set_id(value, id);
+		err = snd_ctl_elem_read(ctl, value);
+		if (err < 0)
+			goto __fail;
+		if (type == SEQUENCE_ELEMENT_TYPE_CSET_BIN_FILE)
+			err = binary_file_parse(value, info, pos);
+		else
+			err = snd_ctl_ascii_value_parse(ctl, value, info, pos);
+		if (err < 0)
+			goto __fail;
+		err = snd_ctl_elem_write(ctl, value);
+		if (err < 0)
+			goto __fail;
+	}
 	err = 0;
       __fail:
 	if (id != NULL)
@@ -207,6 +317,8 @@ static int execute_cset(snd_ctl_t *ctl, const char *cset)
 		free(value);
 	if (info != NULL)
 		free(info);
+	if (res != NULL)
+		free(res);
 
 	return err;
 }
@@ -227,6 +339,7 @@ static int execute_sequence(snd_use_case_mgr_t *uc_mgr,
 	struct sequence_element *s;
 	char *cdev = NULL;
 	snd_ctl_t *ctl = NULL;
+	struct ctl_list *ctl_list;
 	int err = 0;
 
 	list_for_each(pos, seq) {
@@ -238,44 +351,72 @@ static int execute_sequence(snd_use_case_mgr_t *uc_mgr,
 				goto __fail_nomem;
 			break;
 		case SEQUENCE_ELEMENT_TYPE_CSET:
-			if (cdev == NULL) {
-				const char *cdev1 = NULL, *cdev2 = NULL;
-				err = get_value3(&cdev1, "PlaybackCTL",
-						 value_list1,
-						 value_list2,
-						 value_list3);
-				if (err < 0 && err != ENOENT) {
+		case SEQUENCE_ELEMENT_TYPE_CSET_BIN_FILE:
+		case SEQUENCE_ELEMENT_TYPE_CSET_TLV:
+			if (cdev == NULL && uc_mgr->in_component_domain) {
+				/* For sequence of a component device, use
+				 * its parent's cdev stored by ucm manager.
+				 */
+				if (uc_mgr->cdev == NULL) {
 					uc_error("cdev is not defined!");
 					return err;
 				}
-				err = get_value3(&cdev1, "CaptureCTL",
+
+				cdev = strndup(uc_mgr->cdev, PATH_MAX);
+				if (!cdev)
+					return -ENOMEM;
+			} else if (cdev == NULL) {
+				char *playback_ctl = NULL;
+				char *capture_ctl = NULL;
+
+				err = get_value3(uc_mgr, &playback_ctl, "PlaybackCTL",
 						 value_list1,
 						 value_list2,
 						 value_list3);
-				if (err < 0 && err != ENOENT) {
-					free((char *)cdev1);
+				if (err < 0 && err != -ENOENT) {
 					uc_error("cdev is not defined!");
 					return err;
 				}
-				if (cdev1 == NULL || cdev2 == NULL ||
-                                    strcmp(cdev1, cdev2) == 0) {
-					cdev = (char *)cdev1;
-					free((char *)cdev2);
+				err = get_value3(uc_mgr, &capture_ctl, "CaptureCTL",
+						 value_list1,
+						 value_list2,
+						 value_list3);
+				if (err < 0 && err != -ENOENT) {
+					free(playback_ctl);
+					uc_error("cdev is not defined!");
+					return err;
+				}
+				if (playback_ctl == NULL &&
+				    capture_ctl == NULL) {
+					uc_error("cdev is not defined!");
+					return -EINVAL;
+				}
+				if (playback_ctl != NULL &&
+				    capture_ctl != NULL &&
+				    strcmp(playback_ctl, capture_ctl) != 0) {
+					free(playback_ctl);
+					free(capture_ctl);
+					uc_error("cdev is not equal for playback and capture!");
+					return -EINVAL;
+				}
+				if (playback_ctl != NULL) {
+					cdev = playback_ctl;
+					free(capture_ctl);
 				} else {
-					free((char *)cdev1);
-					free((char *)cdev2);
+					cdev = capture_ctl;
 				}
 			}
 			if (ctl == NULL) {
-				err = open_ctl(uc_mgr, &ctl, cdev);
+				err = uc_mgr_open_ctl(uc_mgr, &ctl_list, cdev, 1);
 				if (err < 0) {
 					uc_error("unable to open ctl device '%s'", cdev);
 					goto __fail;
 				}
+				ctl = ctl_list->ctl;
 			}
-			err = execute_cset(ctl, s->data.cset);
+			err = execute_cset(ctl, s->data.cset, s->type);
 			if (err < 0) {
-				uc_error("unable to execute cset '%s'\n", s->data.cset);
+				uc_error("unable to execute cset '%s'", s->data.cset);
 				goto __fail;
 			}
 			break;
@@ -287,6 +428,19 @@ static int execute_sequence(snd_use_case_mgr_t *uc_mgr,
 			if (err < 0)
 				goto __fail;
 			break;
+		case SEQUENCE_ELEMENT_TYPE_CMPT_SEQ:
+			/* Execute enable or disable sequence of a component
+			 * device. Pass the cdev defined by the machine device.
+			 */
+			err = execute_component_seq(uc_mgr,
+						    &s->data.cmpt_seq,
+						    value_list1,
+						    value_list2,
+						    value_list3,
+						    cdev);
+			if (err < 0)
+				goto __fail;
+			break;
 		default:
 			uc_error("unknown sequence command %i", s->type);
 			break;
@@ -302,6 +456,109 @@ static int execute_sequence(snd_use_case_mgr_t *uc_mgr,
 
 }
 
+/* Execute enable or disable sequence of a component device.
+ *
+ * For a component device (a codec or embedded DSP), its sequence doesn't
+ * specify the sound card device 'cdev', because a component can be reused
+ * by different sound cards (machines). So when executing its sequence, a
+ * parameter 'cdev' is used to pass cdev defined by the sequence of its
+ * parent, the machine device. UCM manger will store the cdev when entering
+ * the component domain.
+ */
+static int execute_component_seq(snd_use_case_mgr_t *uc_mgr,
+				 struct component_sequence *cmpt_seq,
+				 struct list_head *value_list1 ATTRIBUTE_UNUSED,
+				 struct list_head *value_list2 ATTRIBUTE_UNUSED,
+				 struct list_head *value_list3 ATTRIBUTE_UNUSED,
+				 char *cdev)
+{
+	struct use_case_device *device = cmpt_seq->device;
+	struct list_head *seq;
+	int err;
+
+	/* enter component domain and store cdev for the component */
+	uc_mgr->in_component_domain = 1;
+	uc_mgr->cdev = cdev;
+
+	/* choose enable or disable sequence of the component device */
+	if (cmpt_seq->enable)
+		seq = &device->enable_list;
+	else
+		seq = &device->disable_list;
+
+	/* excecute the sequence of the component dev */
+	err = execute_sequence(uc_mgr, seq,
+			       &device->value_list,
+			       &uc_mgr->active_verb->value_list,
+			       &uc_mgr->value_list);
+
+	/* exit component domain and clear cdev */
+	uc_mgr->in_component_domain = 0;
+	uc_mgr->cdev = NULL;
+
+	return err;
+}
+
+static int add_auto_value(snd_use_case_mgr_t *uc_mgr, const char *key, char *value)
+{
+	char *s;
+	int err;
+
+	err = get_value1(uc_mgr, &value, &uc_mgr->value_list, key);
+	if (err == -ENOENT) {
+		s = strdup(value);
+		if (s == NULL)
+			return -ENOMEM;
+		return uc_mgr_add_value(&uc_mgr->value_list, key, s);
+	} else if (err < 0) {
+		return err;
+	}
+	free(value);
+	return 0;
+}
+
+static int add_auto_values(snd_use_case_mgr_t *uc_mgr)
+{
+	struct ctl_list *ctl_list;
+	const char *id;
+	char buf[40];
+	int err;
+
+	ctl_list = uc_mgr_get_master_ctl(uc_mgr);
+	if (ctl_list) {
+		id = snd_ctl_card_info_get_id(ctl_list->ctl_info);
+		snprintf(buf, sizeof(buf), "hw:%s", id);
+		err = add_auto_value(uc_mgr, "PlaybackCTL", buf);
+		if (err < 0)
+			return err;
+		err = add_auto_value(uc_mgr, "CaptureCTL", buf);
+		if (err < 0)
+			return err;
+	}
+	return 0;
+}
+
+/**
+ * \brief execute default commands
+ * \param uc_mgr Use case manager
+ * \return zero on success, otherwise a negative error code
+ */
+static int set_defaults(snd_use_case_mgr_t *uc_mgr)
+{
+	int err;
+
+	if (uc_mgr->default_list_executed)
+		return 0;
+	err = execute_sequence(uc_mgr, &uc_mgr->default_list,
+			       &uc_mgr->value_list, NULL, NULL);
+	if (err < 0) {
+		uc_error("Unable to execute default sequence");
+		return err;
+	}
+	uc_mgr->default_list_executed = 1;
+	return 0;
+}
+
 /**
  * \brief Import master config and execute the default sequence
  * \param uc_mgr Use case manager
@@ -314,11 +571,32 @@ static int import_master_config(snd_use_case_mgr_t *uc_mgr)
 	err = uc_mgr_import_master_config(uc_mgr);
 	if (err < 0)
 		return err;
-	err = execute_sequence(uc_mgr, &uc_mgr->default_list,
-			       &uc_mgr->value_list, NULL, NULL);
-	if (err < 0)
-		uc_error("Unable to execute default sequence");
-	return err;
+	return add_auto_values(uc_mgr);
+}
+
+/**
+ * \brief Check, if the UCM configuration is empty
+ * \param uc_mgr Use case Manager
+ * \return zero on success, otherwise a negative error code
+ */
+static int check_empty_configuration(snd_use_case_mgr_t *uc_mgr)
+{
+	int err;
+	char *value;
+
+	err = get_value(uc_mgr, "Linked", &value, NULL, NULL, 1);
+	if (err >= 0) {
+		err = strcasecmp(value, "true") == 0 ||
+		      strcmp(value, "1") == 0;
+		free(value);
+		if (err)
+			return 0;
+	}
+	if (!list_empty(&uc_mgr->verb_list))
+		return 0;
+	if (!list_empty(&uc_mgr->boot_list))
+		return 0;
+	return -ENXIO;
 }
 
 /**
@@ -614,6 +892,9 @@ static int set_verb(snd_use_case_mgr_t *uc_mgr,
 	int err;
 
 	if (enable) {
+		err = set_defaults(uc_mgr);
+		if (err < 0)
+			return err;
 		seq = &verb->enable_list;
 	} else {
 		seq = &verb->disable_list;
@@ -701,42 +982,51 @@ static int set_device(snd_use_case_mgr_t *uc_mgr,
  * \param card_name name of card to open
  * \return zero on success, otherwise a negative error code
  */
-int snd_use_case_mgr_open(snd_use_case_mgr_t **mgr,
+int snd_use_case_mgr_open(snd_use_case_mgr_t **uc_mgr,
 			  const char *card_name)
 {
-	snd_use_case_mgr_t *uc_mgr;
+	snd_use_case_mgr_t *mgr;
 	int err;
 
 	/* create a new UCM */
-	uc_mgr = calloc(1, sizeof(snd_use_case_mgr_t));
-	if (uc_mgr == NULL)
+	mgr = calloc(1, sizeof(snd_use_case_mgr_t));
+	if (mgr == NULL)
 		return -ENOMEM;
-	INIT_LIST_HEAD(&uc_mgr->verb_list);
-	INIT_LIST_HEAD(&uc_mgr->default_list);
-	INIT_LIST_HEAD(&uc_mgr->value_list);
-	INIT_LIST_HEAD(&uc_mgr->active_modifiers);
-	INIT_LIST_HEAD(&uc_mgr->active_devices);
-	pthread_mutex_init(&uc_mgr->mutex, NULL);
+	INIT_LIST_HEAD(&mgr->verb_list);
+	INIT_LIST_HEAD(&mgr->boot_list);
+	INIT_LIST_HEAD(&mgr->default_list);
+	INIT_LIST_HEAD(&mgr->value_list);
+	INIT_LIST_HEAD(&mgr->active_modifiers);
+	INIT_LIST_HEAD(&mgr->active_devices);
+	INIT_LIST_HEAD(&mgr->ctl_list);
+	INIT_LIST_HEAD(&mgr->variable_list);
+	pthread_mutex_init(&mgr->mutex, NULL);
 
-	uc_mgr->card_name = strdup(card_name);
-	if (uc_mgr->card_name == NULL) {
-		free(uc_mgr);
+	mgr->card_name = strdup(card_name);
+	if (mgr->card_name == NULL) {
+		free(mgr);
 		return -ENOMEM;
 	}
 
 	/* get info on use_cases and verify against card */
-	err = import_master_config(uc_mgr);
+	err = import_master_config(mgr);
 	if (err < 0) {
 		uc_error("error: failed to import %s use case configuration %d",
-			card_name, err);
-		goto err;
+			 card_name, err);
+		goto _err;
 	}
 
-	*mgr = uc_mgr;
+	err = check_empty_configuration(mgr);
+	if (err < 0) {
+		uc_error("error: failed to import %s (empty configuration)", card_name);
+		goto _err;
+	}
+
+	*uc_mgr = mgr;
 	return 0;
 
-err:
-	uc_mgr_free(uc_mgr);
+_err:
+	uc_mgr_free(mgr);
 	return err;
 }
 
@@ -753,10 +1043,12 @@ int snd_use_case_mgr_reload(snd_use_case_mgr_t *uc_mgr)
 
 	uc_mgr_free_verb(uc_mgr);
 
+	uc_mgr->default_list_executed = 0;
+
 	/* reload all use cases */
 	err = import_master_config(uc_mgr);
 	if (err < 0) {
-		uc_error("error: failed to reload use cases\n");
+		uc_error("error: failed to reload use cases");
 		pthread_mutex_unlock(&uc_mgr->mutex);
 		return -EINVAL;
 	}
@@ -840,7 +1132,6 @@ int snd_use_case_mgr_reset(snd_use_case_mgr_t *uc_mgr)
 /**
  * \brief Get list of verbs in pair verbname+comment
  * \param list Returned list
- * \param verbname For verb (NULL = current)
  * \return Number of list entries if success, otherwise a negative error code
  */
 static int get_verb_list(snd_use_case_mgr_t *uc_mgr, const char **list[])
@@ -928,8 +1219,10 @@ static int get_supcon_device_list(snd_use_case_mgr_t *uc_mgr,
 
 	modifier = find_modifier(uc_mgr, verb, name, 0);
 	if (modifier) {
-		if (modifier->dev_list.type != type)
+		if (modifier->dev_list.type != type) {
+			*list = NULL;
 			return 0;
+		}
 		return get_list(&modifier->dev_list.list, list,
 				struct dev_list_node, list,
 				name);
@@ -937,15 +1230,16 @@ static int get_supcon_device_list(snd_use_case_mgr_t *uc_mgr,
 
 	device = find_device(uc_mgr, verb, name, 0);
 	if (device) {
-		if (device->dev_list.type != type)
+		if (device->dev_list.type != type) {
+			*list = NULL;
 			return 0;
+		}
 		return get_list(&device->dev_list.list, list,
 				struct dev_list_node, list,
 				name);
 	}
 
 	return -ENOENT;
-
 }
 
 /**
@@ -972,41 +1266,203 @@ static int get_conflicting_device_list(snd_use_case_mgr_t *uc_mgr,
 	return get_supcon_device_list(uc_mgr, list, name, DEVLIST_CONFLICTING);
 }
 
+#ifndef DOC_HIDDEN
 struct myvalue {
-        struct list_head list;
-        char *value;
+	struct list_head list;
+	const char *text;
 };
+#endif
 
+/**
+ * \brief Convert myvalue list string list
+ * \param list myvalue list
+ * \param res string list
+ * \retval Number of list entries if success, otherwise a negativer error code
+ */
+static int myvalue_to_str_list(struct list_head *list, char ***res)
+{
+	struct list_head *pos;
+	struct myvalue *value;
+	char **p;
+	int cnt;
+
+	cnt = alloc_str_list(list, 1, res);
+	if (cnt < 0)
+		return cnt;
+	p = *res;
+	list_for_each(pos, list) {
+		value = list_entry(pos, struct myvalue, list);
+		*p = strdup(value->text);
+		if (*p == NULL) {
+			snd_use_case_free_list((const char **)p, cnt);
+			return -ENOMEM;
+		}
+		p++;
+	}
+	return cnt;
+}
+
+/**
+ * \brief Free myvalue list
+ * \param list myvalue list
+ */
+static void myvalue_list_free(struct list_head *list)
+{
+	struct list_head *pos, *npos;
+	struct myvalue *value;
+
+	list_for_each_safe(pos, npos, list) {
+		value = list_entry(pos, struct myvalue, list);
+		list_del(&value->list);
+		free(value);
+	}
+}
+
+/**
+ * \brief Merge one value to the myvalue list
+ * \param list The list with values
+ * \param value The value to be merged (without duplicates)
+ * \return 1 if dup, 0 if success, otherwise a negative error code
+ */
+static int merge_value(struct list_head *list, const char *text)
+{
+	struct list_head *pos;
+	struct myvalue *value;
+
+	list_for_each(pos, list) {
+		value = list_entry(pos, struct myvalue, list);
+		if (strcmp(value->text, text) == 0)
+			return 1;
+	}
+	value = malloc(sizeof(*value));
+	if (value == NULL)
+		return -ENOMEM;
+	value->text = text;
+	list_add_tail(&value->list, list);
+	return 0;
+}
+
+/**
+ * \brief Find all values for given identifier
+ * \param list Returned list
+ * \param source Source list with ucm_value structures
+ * \return Zero if success, otherwise a negative error code
+ */
+static int add_identifiers(struct list_head *list,
+			   struct list_head *source)
+{
+	struct ucm_value *v;
+	struct list_head *pos;
+	int err;
+
+	list_for_each(pos, source) {
+		v = list_entry(pos, struct ucm_value, list);
+		err = merge_value(list, v->name);
+		if (err < 0)
+			return err;
+	}
+	return 0;
+}
+
+/**
+ * \brief Find all values for given identifier
+ * \param list Returned list
+ * \param identifier Identifier
+ * \param source Source list with ucm_value structures
+ */
 static int add_values(struct list_head *list,
                       const char *identifier,
                       struct list_head *source)
 {
-        struct ucm_value *v;
-        struct myvalue *val;
-        struct list_head *pos, *pos1;
-        int match;
+	struct ucm_value *v;
+	struct list_head *pos;
+	int err;
         
-        list_for_each(pos, source) {
-                v = list_entry(pos, struct ucm_value, list);
-                if (check_identifier(identifier, v->name)) {
-                        match = 0;
-                        list_for_each(pos1, list) {
-                                val = list_entry(pos1, struct myvalue, list);
-                                if (strcmp(val->value, v->data) == 0) {
-                                        match = 1;
-                                        break;
-                                }
-                        }
-                        if (!match) {
-                                val = malloc(sizeof(struct myvalue));
-                                if (val == NULL)
-                                        return -ENOMEM;
-				val->value = v->data;
-                                list_add_tail(&val->list, list);
-                        }
-                }
-        }
-        return 0;
+	list_for_each(pos, source) {
+		v = list_entry(pos, struct ucm_value, list);
+		if (check_identifier(identifier, v->name)) {
+			err = merge_value(list, v->data);
+			if (err < 0)
+				return err;
+		}
+	}
+	return 0;
+}
+
+/**
+ * \brief compare two identifiers
+ */
+static int identifier_cmp(const void *_a, const void *_b)
+{
+	const char * const *a = _a;
+	const char * const *b = _b;
+	return strcmp(*a, *b);
+}
+
+/**
+ * \brief Get list of available identifiers
+ * \param list Returned list
+ * \param name Name of verb or modifier to query
+ * \return Number of list entries if success, otherwise a negative error code
+ */
+static int get_identifiers_list(snd_use_case_mgr_t *uc_mgr,
+				const char **list[], char *name)
+{
+	struct use_case_verb *verb;
+	struct use_case_modifier *modifier;
+	struct use_case_device *device;
+	struct list_head mylist;
+	struct list_head *value_list;
+	char *str, **res;
+	int err;
+
+	if (!name)
+		return -ENOENT;
+
+	str = strchr(name, '/');
+	if (str) {
+		*str = '\0';
+		verb = find_verb(uc_mgr, str + 1);
+	}
+	else {
+		verb = uc_mgr->active_verb;
+	}
+	if (!verb)
+		return -ENOENT;
+
+	value_list = NULL;
+	modifier = find_modifier(uc_mgr, verb, name, 0);
+	if (modifier) {
+		value_list = &modifier->value_list;
+	} else {
+		device = find_device(uc_mgr, verb, name, 0);
+		if (device)
+			value_list = &device->value_list;
+	}
+	if (value_list == NULL)
+		return -ENOENT;
+
+	INIT_LIST_HEAD(&mylist);
+	err = add_identifiers(&mylist, &uc_mgr->value_list);
+	if (err < 0)
+		goto __fail;
+	err = add_identifiers(&mylist, &verb->value_list);
+	if (err < 0)
+		goto __fail;
+	err = add_identifiers(&mylist, value_list);
+	if (err < 0)
+		goto __fail;
+	err = myvalue_to_str_list(&mylist, &res);
+	if (err > 0)
+		*list = (const char **)res;
+	else if (err == 0)
+		*list = NULL;
+__fail:
+	myvalue_list_free(&mylist);
+	if (err <= 0)
+		return err;
+	qsort(*list, err, sizeof(char *), identifier_cmp);
+	return err;
 }
 
 /**
@@ -1020,8 +1476,7 @@ static int get_value_list(snd_use_case_mgr_t *uc_mgr,
                           const char **list[],
                           char *verbname)
 {
-        struct list_head mylist, *pos, *npos;
-        struct myvalue *val;
+	struct list_head mylist, *pos;
         struct use_case_verb *verb;
         struct use_case_device *dev;
         struct use_case_modifier *mod;
@@ -1054,26 +1509,13 @@ static int get_value_list(snd_use_case_mgr_t *uc_mgr,
                 if (err < 0)
                         goto __fail;
         }
-        err = alloc_str_list(&mylist, 1, &res);
-        if (err >= 0) {
+	err = myvalue_to_str_list(&mylist, &res);
+	if (err > 0)
 	        *list = (const char **)res;
-                list_for_each(pos, &mylist) {
-                        val = list_entry(pos, struct myvalue, list);
-                        *res = strdup(val->value);
-                        if (*res == NULL) {
-                                snd_use_case_free_list((const char **)res, err);
-                                err = -ENOMEM;
-                                goto __fail;
-                        }
-                        res++;
-                }
-        }
+	else if (err == 0)
+		*list = NULL;
       __fail:
-        list_for_each_safe(pos, npos, &mylist) {
-                val = list_entry(pos, struct myvalue, list);
-                list_del(&val->list);
-                free(val);
-        }
+	myvalue_list_free(&mylist);
         return err;
 }
 
@@ -1143,28 +1585,30 @@ int snd_use_case_get_list(snd_use_case_mgr_t *uc_mgr,
                 } else {
                         str = NULL;
                 }
-        	if (check_identifier(identifier, "_devices"))
-          		err = get_device_list(uc_mgr, list, str);
+		if (check_identifier(identifier, "_devices"))
+			err = get_device_list(uc_mgr, list, str);
                 else if (check_identifier(identifier, "_modifiers"))
-                        err = get_modifier_list(uc_mgr, list, str);
-                else if (check_identifier(identifier, "_supporteddevs"))
-                        err = get_supported_device_list(uc_mgr, list, str);
-                else if (check_identifier(identifier, "_conflictingdevs"))
-                        err = get_conflicting_device_list(uc_mgr, list, str);
+			err = get_modifier_list(uc_mgr, list, str);
+		else if (check_identifier(identifier, "_identifiers"))
+			err = get_identifiers_list(uc_mgr, list, str);
+		else if (check_identifier(identifier, "_supporteddevs"))
+			err = get_supported_device_list(uc_mgr, list, str);
+		else if (check_identifier(identifier, "_conflictingdevs"))
+			err = get_conflicting_device_list(uc_mgr, list, str);
 		else if (identifier[0] == '_')
 			err = -ENOENT;
-                else
-                        err = get_value_list(uc_mgr, identifier, list, str);
-        	if (str)
-        		free(str);
-        }
+		else
+			err = get_value_list(uc_mgr, identifier, list, str);
+		if (str)
+			free(str);
+	}
       __end:
 	pthread_mutex_unlock(&uc_mgr->mutex);
 	return err;
 }
 
-static int get_value1(const char **value, struct list_head *value_list,
-                      const char *identifier)
+static int get_value1(snd_use_case_mgr_t *uc_mgr, char **value,
+		      struct list_head *value_list, const char *identifier)
 {
         struct ucm_value *val;
         struct list_head *pos;
@@ -1173,18 +1617,22 @@ static int get_value1(const char **value, struct list_head *value_list,
 		return -ENOENT;
 
         list_for_each(pos, value_list) {
-              val = list_entry(pos, struct ucm_value, list);
-              if (check_identifier(identifier, val->name)) {
-                      *value = strdup(val->data);
-                      if (*value == NULL)
-                              return -ENOMEM;
-                      return 0;
-              }
+		val = list_entry(pos, struct ucm_value, list);
+		if (check_identifier(identifier, val->name)) {
+			if (uc_mgr->conf_format < 2) {
+				*value = strdup(val->data);
+				if (*value == NULL)
+					return -ENOMEM;
+				return 0;
+			}
+			return uc_mgr_get_substituted_value(uc_mgr, value, val->data);
+		}
         }
         return -ENOENT;
 }
 
-static int get_value3(const char **value,
+static int get_value3(snd_use_case_mgr_t *uc_mgr,
+		      char **value,
 		      const char *identifier,
 		      struct list_head *value_list1,
 		      struct list_head *value_list2,
@@ -1192,13 +1640,13 @@ static int get_value3(const char **value,
 {
 	int err;
 
-	err = get_value1(value, value_list1, identifier);
+	err = get_value1(uc_mgr, value, value_list1, identifier);
 	if (err >= 0 || err != -ENOENT)
 		return err;
-	err = get_value1(value, value_list2, identifier);
+	err = get_value1(uc_mgr, value, value_list2, identifier);
 	if (err >= 0 || err != -ENOENT)
 		return err;
-	err = get_value1(value, value_list3, identifier);
+	err = get_value1(uc_mgr, value, value_list3, identifier);
 	if (err >= 0 || err != -ENOENT)
 		return err;
 	return -ENOENT;
@@ -1214,7 +1662,7 @@ static int get_value3(const char **value,
  */
 static int get_value(snd_use_case_mgr_t *uc_mgr,
 			const char *identifier,
-			const char **value,
+			char **value,
 			const char *mod_dev_name,
 			const char *verb_name,
 			int exact)
@@ -1235,7 +1683,7 @@ static int get_value(snd_use_case_mgr_t *uc_mgr,
 				mod = find_modifier(uc_mgr, verb,
 						    mod_dev_name, 0);
 				if (mod) {
-					err = get_value1(value,
+					err = get_value1(uc_mgr, value,
 							 &mod->value_list,
 							 identifier);
 					if (err >= 0 || err != -ENOENT)
@@ -1245,7 +1693,7 @@ static int get_value(snd_use_case_mgr_t *uc_mgr,
 				dev = find_device(uc_mgr, verb,
 						  mod_dev_name, 0);
 				if (dev) {
-					err = get_value1(value,
+					err = get_value1(uc_mgr, value,
 							 &dev->value_list,
 							 identifier);
 					if (err >= 0 || err != -ENOENT)
@@ -1256,7 +1704,7 @@ static int get_value(snd_use_case_mgr_t *uc_mgr,
 					return -ENOENT;
 			}
 
-			err = get_value1(value, &verb->value_list, identifier);
+			err = get_value1(uc_mgr, value, &verb->value_list, identifier);
 			if (err >= 0 || err != -ENOENT)
 				return err;
 		}
@@ -1265,7 +1713,7 @@ static int get_value(snd_use_case_mgr_t *uc_mgr,
 			return -ENOENT;
 	}
 
-	err = get_value1(value, &uc_mgr->value_list, identifier);
+	err = get_value1(uc_mgr, value, &uc_mgr->value_list, identifier);
 	if (err >= 0 || err != -ENOENT)
 		return err;
 
@@ -1310,6 +1758,21 @@ int snd_use_case_get(snd_use_case_mgr_t *uc_mgr,
                         goto __end;
                 }
 	        err = 0;
+	} else if (strcmp(identifier, "_file") == 0) {
+		/* get the conf file name of the opened card */
+		if ((uc_mgr->card_name == NULL) ||
+		    (uc_mgr->conf_file_name == NULL) ||
+		    (uc_mgr->conf_file_name[0] == '\0')) {
+			err = -ENOENT;
+			goto __end;
+		}
+		*value = strdup(uc_mgr->conf_file_name);
+		if (*value == NULL) {
+			err = -ENOMEM;
+			goto __end;
+		}
+		err = 0;
+
 	} else if (identifier[0] == '_') {
 		err = -ENOENT;
 		goto __end;
@@ -1345,7 +1808,8 @@ int snd_use_case_get(snd_use_case_mgr_t *uc_mgr,
 			verb = NULL;
 		}
 
-		err = get_value(uc_mgr, ident, value, mod_dev, verb, exact);
+		err = get_value(uc_mgr, ident, (char **)value, mod_dev, verb,
+		                exact);
 		if (ident != identifier)
 			free((void *)ident);
 		if (mod_dev)
@@ -1385,12 +1849,20 @@ int snd_use_case_geti(snd_use_case_mgr_t *uc_mgr,
                         str = NULL;
                 }
                 if (check_identifier(identifier, "_devstatus")) {
+			if (!str) {
+				err = -EINVAL;
+				goto __end;
+			}
                         err = device_status(uc_mgr, str);
 			if (err >= 0) {
 				*value = err;
 				err = 0;
 			}
 		} else if (check_identifier(identifier, "_modstatus")) {
+			if (!str) {
+				err = -EINVAL;
+				goto __end;
+			}
                         err = modifier_status(uc_mgr, str);
 			if (err >= 0) {
 				*value = err;
@@ -1414,6 +1886,34 @@ int snd_use_case_geti(snd_use_case_mgr_t *uc_mgr,
         return err;
 }
 
+static int set_boot_user(snd_use_case_mgr_t *uc_mgr,
+			 const char *value)
+{
+	int err;
+
+	if (value != NULL && *value) {
+		uc_error("error: wrong value for _boot (%s)", value);
+		return -EINVAL;
+	}
+	err = execute_sequence(uc_mgr, &uc_mgr->boot_list,
+			       &uc_mgr->value_list, NULL, NULL);
+	if (err < 0) {
+		uc_error("Unable to execute boot sequence");
+		return err;
+	}
+	return err;
+}
+
+static int set_defaults_user(snd_use_case_mgr_t *uc_mgr,
+			     const char *value)
+{
+	if (value != NULL && *value) {
+		uc_error("error: wrong value for _defaults (%s)", value);
+		return -EINVAL;
+	}
+	return set_defaults(uc_mgr);
+}
+
 static int handle_transition_verb(snd_use_case_mgr_t *uc_mgr,
                                   struct use_case_verb *new_verb)
 {
@@ -1440,7 +1940,7 @@ static int set_verb_user(snd_use_case_mgr_t *uc_mgr,
                          const char *verb_name)
 {
         struct use_case_verb *verb;
-        int err;
+        int err = 0;
 
         if (uc_mgr->active_verb &&
             strcmp(uc_mgr->active_verb->name, verb_name) == 0)
@@ -1622,10 +2122,14 @@ int snd_use_case_set(snd_use_case_mgr_t *uc_mgr,
                      const char *value)
 {
 	char *str, *str1;
-	int err;
+	int err = 0;
 
 	pthread_mutex_lock(&uc_mgr->mutex);
-	if (strcmp(identifier, "_verb") == 0)
+	if (strcmp(identifier, "_boot") == 0)
+		err = set_boot_user(uc_mgr, value);
+	else if (strcmp(identifier, "_defaults") == 0)
+		err = set_defaults_user(uc_mgr, value);
+	else if (strcmp(identifier, "_verb") == 0)
 	        err = set_verb_user(uc_mgr, value);
         else if (strcmp(identifier, "_enadev") == 0)
                 err = set_device_user(uc_mgr, value, 1);
@@ -1644,7 +2148,8 @@ int snd_use_case_set(snd_use_case_mgr_t *uc_mgr,
                 		goto __end;
                         }
                 } else {
-                        str = NULL;
+                        err = -EINVAL;
+                        goto __end;
                 }
                 if (check_identifier(identifier, "_swdev"))
                         err = switch_device(uc_mgr, str, value);
@@ -1659,3 +2164,54 @@ int snd_use_case_set(snd_use_case_mgr_t *uc_mgr,
 	pthread_mutex_unlock(&uc_mgr->mutex);
         return err;
 }
+
+/**
+ * \brief Parse control element identifier
+ * \param elem_id Element identifier
+ * \param ucm_id Use case identifier
+ * \param value String value to be parsed
+ * \return Zero if success, otherwise a negative error code
+ */
+int snd_use_case_parse_ctl_elem_id(snd_ctl_elem_id_t *dst,
+				   const char *ucm_id,
+				   const char *value)
+{
+	snd_ctl_elem_iface_t iface;
+	int jack_control;
+
+	jack_control = strcmp(ucm_id, "JackControl") == 0;
+	if (!jack_control &&
+	    strcmp(ucm_id, "PlaybackVolume") &&
+	    strcmp(ucm_id, "PlaybackSwitch") &&
+	    strcmp(ucm_id, "CaptureVolume") &&
+	    strcmp(ucm_id, "CaptureSwitch"))
+		return -EINVAL;
+	snd_ctl_elem_id_clear(dst);
+	if (strcasestr(ucm_id, "name="))
+		return __snd_ctl_ascii_elem_id_parse(dst, value, NULL);
+	iface = SND_CTL_ELEM_IFACE_MIXER;
+	if (jack_control)
+		iface = SND_CTL_ELEM_IFACE_CARD;
+	snd_ctl_elem_id_set_interface(dst, iface);
+	snd_ctl_elem_id_set_name(dst, value);
+	return 0;
+}
+
+/**
+ * \brief Parse mixer element identifier
+ * \param dst Simple mixer element identifier
+ * \param ucm_id Use case identifier
+ * \param value String value to be parsed
+ * \return Zero if success, otherwise a negative error code
+ */
+int snd_use_case_parse_selem_id(snd_mixer_selem_id_t *dst,
+				const char *ucm_id,
+				const char *value)
+{
+#ifdef BUILD_MIXER
+	if (strcmp(ucm_id, "PlaybackMixerId") == 0 ||
+	    strcmp(ucm_id, "CaptureMixerId") == 0)
+		return snd_mixer_selem_id_parse(dst, value);
+#endif
+	return -EINVAL;
+}
